from dataclasses import dataclass, field, asdict
from collections.abc import Callable
from abc import ABC, abstractmethod
import pandas as pd
import traceback
import inspect

@dataclass
class SimpleTestExport:
    title: str
    function: str
    input: str
    expected_output: str


@dataclass
class SimpleTest:
    title: str
    function: Callable
    input: pd.DataFrame | pd.Series
    expected_output: pd.DataFrame | pd.Series

    def export(self):
        return SimpleTestExport(
            title=self.title,
            function=inspect.getsource(self.function),
            input=self.input.to_csv(index=False),
            expected_output=self.expected_output.to_csv(index=False)
        )

    def __radd__(self, tests: list):
        if isinstance(tests, list):
            tests.append(self)
            return tests
        else:
            raise TypeError(
                (
                    "presyence SimpleTest can only be added to a list, "
                    f"found:, {type(tests)}"
                )
            )

    def run(self):
        try:
            testrun_output = self.function(self.input)
            if isinstance(self.expected_output, pd.DataFrame):
                assert_same = pd.testing.assert_frame_equal
            elif isinstance(self.expected_output, pd.Series):
                assert_same = pd.testing.assert_series_equal
            else:
                raise NotImplementedError("Expected output must be a pandas DataFrame or a Serie")
            if not isinstance(testrun_output, pd.DataFrame | pd.Series):
                raise NotImplementedError
            assert_same(self.expected_output, testrun_output)
        except NotImplementedError:
            raise
        except AssertionError as e:
            result = WrongResult(
                testrun_output = testrun_output.to_csv(index=False),
                assertion_error_message = e.args[0]
            )
        except Exception:
            result = Crash(
                traceback = traceback.format_exc()
            )
        else:
            result = Success()
        finally:
            _test_result = SimpleTestReport(
                test = self.export(),
                result = result
            )
            return asdict(_test_result)


@dataclass
class TestResult(ABC):
    successful: bool = field(init=False)
    status: str = field(init=False)

    @abstractmethod
    def __bool__(self):
        ...

    def __post_init__(self):
        self.successful = bool(self)
        self.status = self.__class__.__name__


@dataclass
class Success(TestResult):
    def __bool__(self):
        return True

@dataclass
class WrongResult(TestResult):
    testrun_output: str
    assertion_error_message: str
    def __bool__(self):
        return False

@dataclass
class Crash(TestResult):
    traceback: str
    def __bool__(self):
        return False

@dataclass
class SimpleTestReport:
    test: SimpleTestExport
    result: TestResult



import typer
import dataclasses
from .testrunner.discovery import discover_test_files, extract_tests_from_file
from .reporter.server import run_server_and_inject_payload

app = typer.Typer()


@app.command()
def main():
    print("Looking for tests")
    test_files = discover_test_files()
    print(test_files)
    if not test_files:
        print("No test file found. Test file names must end with '.spec.sy.py'")
        exit()
    tests = []
    for test_file in test_files:
        _tests = extract_tests_from_file(test_file)
        if not _tests:
            print(f"No tests found in file {test_file.absolute()}")
            exit()
        tests += _tests
    print("Running tests")
    test_results = []
    for test in tests:
        test_results += [test.run()]
    print("Creating the test report")
    test_report = {"testReport": test_results}
    print("Serving the test report on http://localhost:5000/")
    run_server_and_inject_payload(test_report)



""" This file is about looking for test files
"""
from pathlib import Path
import importlib.machinery
import importlib.util
import sys
import types


def discover_test_files(pattern="**/*.spec.sy.py", directory="."):
    p = Path(directory)
    test_files = p.glob(pattern)
    return list(
        test_files
    )  # return a list of files instead of an iterator for convenience


def extract_tests_from_file(file):
    # get the true absolute name of the file
    file = file.resolve()
    parent, name = file.parent, file.name
    # add parent to path to help with relative imports 
    if parent not in sys.path:
        sys.path.append(str(parent))
    # in the next lines we will import the file
    module_name = name
    spec = importlib.util.spec_from_file_location(module_name, file)
    # test_module is the .py containing the tests to run 
    test_module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = test_module
    spec.loader.exec_module(test_module)
    # check that a lists named TESTS is present
    tests = test_module.__dict__.get("TESTS", [])
    return tests



from .access_static import get_website_files
from .inject_json_into_js import inject_payload
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse, HTMLResponse, Response


def run_server_and_inject_payload(payload):
    static_files = get_website_files()
    static_files = { f.route : f for f in static_files }

    index_file = static_files.pop("/index.html")

    worker_route = [k for k in static_files.keys() if k.startswith("/assets/worker")][0]
    worker_file = static_files[worker_route]

    worker_file.content = inject_payload(worker_file.content, payload)

    async def catch_all(request, exc):
        return HTMLResponse(index_file.content, status_code=200)

    exceptions = {
        404: catch_all,
    }

    app = FastAPI(exception_handlers=exceptions)

    @app.get("/")
    def home():
        return HTMLResponse(index_file.content, status_code=200)

    def make_route(static_file):
        extension = static_file.route.split('.')[-1]
        common_media_types = {
            "js" : "application/javascript",
            "ico" : "image/x-icon",
        }
        mime_type = common_media_types.get(extension, "text/html")
        @app.get(static_file.route)
        def route():
            return Response(static_file.content, media_type=mime_type)

    for each in static_files.values():
        make_route(each)

    uvicorn.run(app, port=5000, log_level="info")



from dataclasses import dataclass, field
import importlib.resources as resources
from typing import TextIO

@dataclass
class StaticFile:
    package_name: str
    resource_name: str
    content: str | bytes = field(init=False)
    def __post_init__(self):
        try:
            self.content = resources.read_text(self.package_name, self.resource_name)
        except UnicodeDecodeError:
            self.content = resources.read_binary(self.package_name, self.resource_name)

@dataclass
class WebsiteStaticFile(StaticFile):
    @property
    def route(self) -> str:
        """ URL route corresponding to that static file"""
        parent = "/".join(self.package_name.split('www')[-1].split('.'))
        return "/".join([parent, self.resource_name])

def get_website_static_files_from_folder(package_name):
    result = []
    folder_files = resources.contents(package_name)
    for file_name in folder_files:
        if resources.is_resource(package_name, file_name):
            if not file_name.startswith("__"):
                result.append(
                    WebsiteStaticFile(
                        package_name,
                        file_name
                    )
                )
    return result

def get_website_files() -> list[WebsiteStaticFile]:
    static_files = []
    for package_name in ["presyence.reporter.www", "presyence.reporter.www.assets"]:
        static_files += get_website_static_files_from_folder(package_name)
    return static_files



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['presyence',
 'presyence.reporter',
 'presyence.reporter.www',
 'presyence.reporter.www.assets',
 'presyence.testrunner']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.80.0,<0.81.0',
 'pandas>=1.4.3,<2.0.0',
 'typer>=0.6.1,<0.7.0',
 'uvicorn>=0.18.3,<0.19.0']

entry_points = \
{'console_scripts': ['presyence = presyence.main:app']}

setup_kwargs = {
    'name': 'presyence',
    'version': '0.1.8',
    'description': '',
    'long_description': None,
    'author': 'Rémi Connesson',
    'author_email': 'remiconnesson@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import json


def inject_payload(content, payload):
    # insert a place holder easier to replace
    begin_payload, end_payload = "beginPayload:!0,", ",endPayload:!0"
    start_placeholder = content.index(begin_payload) + len(begin_payload)
    end_placeholder = content.index(end_payload)
    content = list(content)
    new_placeholder = "REPLACEMEPLEASE"
    content[start_placeholder:end_placeholder] = list(new_placeholder)
    content = "".join(content)

    new_payload = json.dumps(
        payload, separators=(",", ":")
    )  # separators with out whitespace
    trim_first_and_last_character = lambda s: s[1:-1]  # remove `{` and `}`
    new_payload = trim_first_and_last_character(new_payload)

    # inject the string in place of the placeholder
    content = content.replace(new_placeholder, new_payload)

    return content
